#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) Pilot Systems and Libération, 2010-2013

# This file is part of SeSQL.

# SeSQL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.

# SeSQL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with SeSQL.  If not, see <http://www.gnu.org/licenses/>.

"""
Update the scheduled updates
"""
import sys, os, time

import traceback
import logging

import daemon
import lockfile
from optparse import OptionParser

from sesql import index
from sesql import config
from sesql import results
from sesql import __version__

from django.db import connection, transaction

logger = logging.getLogger('sesql-update')

def version():
    print "sesql update daemon, v " + __version__

class UpdateDaemon(object):
    """
    The daemon class
    """
    def __init__(self, options):
        """
        Constructor
        """
        self.options = options

    def start_deamon(self):
        """
        Start the daemon
        """
        if self.options.debug:
            self.setup_logging()
            return self.run()

        with daemon.DaemonContext():            
            self.setup_logging()
            self.write_pid()
            self.run()

    def setup_logging(self):
        """
        Setup the logging
        """
        if self.options.debug:
            ch = logging.StreamHandler()
            ch.setLevel(logging.DEBUG)
            logger.setLevel(logging.DEBUG)
        elif self.options.logfile:
            ch = logging.FileHandler(self.options.logfile)
            ch.setLevel(logging.INFO)
            logger.setLevel(logging.INFO)
        else:
            return
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        ch.setFormatter(formatter)
        logger.addHandler(ch)

    def write_pid(self):
        """
        Write the PID file
        """
        open(self.options.pidfile, 'w').write(str(os.getpid()))

    def run(self):
        """
        Main loop
        """
        while True:
            try:
                self.process_chunk()
            except:
                type, value, tb = sys.exc_info()
                error = traceback.format_exception_only(type, value)[0]
                logger.error(error)
            connection.close()
            logger.debug("Sleeping for %.2f second(s)" % self.options.delay)
            time.sleep(self.options.delay)

    @transaction.commit_manually
    def process_chunk(self):
        """
        Process a chunk
        """
        QUERY = """SELECT classname, objid, from_related
        FROM sesql_reindex_schedule
        ORDER BY from_related ASC, scheduled_at ASC LIMIT %d"""
        cursor = connection.cursor()
        cursor.execute(QUERY % self.options.chunk)
        rows = cursor.fetchall()
        if not rows:
            transaction.rollback()
            return
        logger.info("Found %d row(s) to reindex" % len(rows))

        done = set()
        cascade = getattr(config, "DAEMON_CASCADE_RELATED", True)

        for row in rows:
            try:
                from_related = row[2]
                row = tuple(row[:2])
                if not row in done:
                    index_related = cascade or not from_related
                    logger.info("Reindexing %s:%d %s" % (row[0], row[1],
                                                         index_related and "with related" or ""))
                    done.add(row)
                    try:
                        obj = results.SeSQLResultSet.load(row)
                        index.index(obj, index_related = index_related)
                    except config.orm.not_found:
                        logger.info("%s:%d doesn't exist anymore, unindexing" % row)
                        index.unindex(row)
                    cursor.execute("""DELETE FROM sesql_reindex_schedule
                                      WHERE classname=%s AND objid=%s""", row)
            except Exception, e:
                logger.error('Error in row %s:%s : %s' % (row[0], row[1], e))
                if self.options.debug:
                    import pdb
                    pdb.post_mortem()
        transaction.commit()

if __name__ == "__main__":
    parser = OptionParser(usage = "usage: %prog [options]")
    parser.add_option('-d', '--debug', action="store_true", dest="debug", 
                      default=False,
                      help="Run in debug mode (don't daemonize)")
    parser.add_option('-V', '--version', action="store_true", dest="version", 
                      default=False,
                      help="Show version number of exit")
    parser.add_option('-c', '--chunk', type="int", dest="chunk", 
                      default = config.DAEMON_DEFAULT_CHUNK,
                      help="Chunk size")
    parser.add_option('-w', '--wait', type="float", dest="delay", 
                      default = config.DAEMON_DEFAULT_DELAY,
                      help="Wait between each chunk (seconds)")
    parser.add_option('-p', '--pidfile', dest="pidfile",
                      default = config.DAEMON_DEFAULT_PID, 
                      help="Pidfile to use")
    parser.add_option('-l', '--logfile', dest="logfile",
                      default = getattr(config, 'DAEMON_DEFAULT_LOG', None), 
                      help="Logfile to use")
    
    (options, args) = parser.parse_args()

    if options.version:
        version()
        sys.exit(0)

    ud = UpdateDaemon(options)
    ud.start_deamon()
