# -*- coding: utf-8 -*-

# Copyright (c) Pilot Systems and Libération, 2016

# This file is part of SeSQL.

# SeSQL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.

# SeSQL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with SeSQL.  If not, see <http://www.gnu.org/licenses/>.

import thread

class UpdateQueueMiddleware(object):
    """
    Middleware that allows queuing callbacks (ie, reindexing) which
    will be done at the end of the request, but only once if queued
    several times
    """
    _queues = {}
    enabled = False

    def __init__(self):
        """
        Mark the middleware is being enabled when it's called
        """
        self.__class__.enabled = True
    
    def process_request(self, request):
        """
        Create an empty queue for that request
        """
        self.init_queue()

    def process_response(self, request, response):
        """
        Process the queue and then delete it
        """
        self.process_queue()
        self.delete_queue()
        return response

    def process_exception(self, request, exception):
        """
        Delete the queue
        """
        self.delete_queue()

    @classmethod
    def get_queue(cls):
        """
        Retrieve queue
        """
        return cls._queues.get(thread.get_ident(), {})

    @classmethod
    def init_queue(cls):
        """
        Create new empty queue
        """
        cls._queues[thread.get_ident()] = {}

    @classmethod
    def delete_queue(cls):
        """
        Delete request
        """
        cls._queues.pop(thread.get_ident(), None)

    @classmethod
    def process_queue(cls):
        """
        Process all events in the queue
        """
        queue = cls.get_queue()
        events = tuple(queue.values())
        for callback, args, kwargs in events:
            callback(*args, **kwargs)

    @classmethod
    def add_to_queue(cls, key, callback, *args, **kwargs):
        """
        Add a callback to the queue
        
        If the middleware isn't enabled, then the callback is called directly
        instead

        The key must be hashable and unique, if the same key is given twice, only
        the last callback will be called
        """
        if not cls.enabled:
            callback(*args, **kwargs)
        else:
            queue = cls.get_queue()
            queue[key] = (callback, args, kwargs)
